/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, 2013, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * Author: Eitan Marder-Eppstein
 *         David V. Lu!!
 *********************************************************************/
#include <iri_opendrive_global_planner/opendrive_planner.h>
#include <pluginlib/class_list_macros.h>
#include <costmap_2d/cost_values.h>
#include <costmap_2d/costmap_2d.h>
#include <tf2/utils.h>

#include "exceptions.h"

//register this planner as a BaseOpendriveGlobalPlanner plugin
PLUGINLIB_EXPORT_CLASS(iri_opendrive_global_planner::OpendriveGlobalPlanner, nav_core::BaseGlobalPlanner)

namespace iri_opendrive_global_planner {

OpendriveGlobalPlanner::OpendriveGlobalPlanner() :
  costmap_(NULL), initialized_(false) {
  this->angle_tol=DEFAULT_ANGLE_THRESHOLD;
  this->dist_tol=DEFAULT_DIST_THRESHOLD;
  this->multi_hyp=false;
}

OpendriveGlobalPlanner::OpendriveGlobalPlanner(std::string name, costmap_2d::Costmap2D* costmap, std::string frame_id) :
  OpendriveGlobalPlanner() 
{
  //initialize the planner
  initialize(name, costmap, frame_id);
}

OpendriveGlobalPlanner::~OpendriveGlobalPlanner() 
{
  if (dsrv_)
    delete dsrv_;
}

void OpendriveGlobalPlanner::initialize(std::string name, costmap_2d::Costmap2DROS* costmap_ros) 
{
  initialize(name, costmap_ros->getCostmap(), costmap_ros->getGlobalFrameID());
}

void OpendriveGlobalPlanner::initialize(std::string name, costmap_2d::Costmap2D* costmap, std::string frame_id) 
{
  double value;

  if (!initialized_) 
  {
    ros::NodeHandle private_nh("~/" + name);
    costmap_ = costmap;
    frame_id_ = frame_id;

    try{
      std::string opendrive_file;
      private_nh.param("resolution", value,0.0);
      this->roadmap.set_resolution(value);
      private_nh.param("scale_factor", value,0.0);
      this->roadmap.set_scale_factor(value);
      private_nh.param("min_road_length", value,0.0);
      this->roadmap.set_min_road_length(value);
      private_nh.param("opendrive_file", opendrive_file,std::string(""));
      if(opendrive_file!="")
        this->roadmap.load(opendrive_file);
      private_nh.param("angle_tol", this->angle_tol,DEFAULT_ANGLE_THRESHOLD);
      private_nh.param("dist_tol", this->dist_tol,DEFAULT_DIST_THRESHOLD);
      private_nh.param("multi_hyp", this->multi_hyp,false);
      std::cout << "multi hypothessis: " << this->multi_hyp << std::endl;
      int cost_type;
      private_nh.param("cost_type",cost_type,(int)COST_DIST);
      this->roadmap.set_cost_type((cost_t)cost_type);
    }catch(CException &e){
      ROS_ERROR_STREAM(e.what());
    }

    plan_pub_ = private_nh.advertise<nav_msgs::Path>("plan", 1);
    make_plan_srv_ = private_nh.advertiseService("make_plan", &OpendriveGlobalPlanner::makePlanService, this);

    dsrv_ = new dynamic_reconfigure::Server<iri_opendrive_global_planner::OpendriveGlobalPlannerConfig>(ros::NodeHandle("~/" + name));
    dynamic_reconfigure::Server<iri_opendrive_global_planner::OpendriveGlobalPlannerConfig>::CallbackType cb = boost::bind(&OpendriveGlobalPlanner::reconfigureCB, this, _1, _2);
    dsrv_->setCallback(cb);

    initialized_ = true;
  } else
    ROS_WARN("This planner has already been initialized, you can't call it twice, doing nothing");
}

void OpendriveGlobalPlanner::reconfigureCB(iri_opendrive_global_planner::OpendriveGlobalPlannerConfig& new_config, uint32_t level) 
{
  try{
    this->roadmap.set_resolution(new_config.resolution);
    this->roadmap.set_scale_factor(new_config.scale_factor);
    this->roadmap.set_min_road_length(new_config.min_road_length);
    if(new_config.opendrive_file!=this->config.opendrive_file || new_config.scale_factor!=this->config.scale_factor || new_config.min_road_length!=this->config.min_road_length)
      this->roadmap.load(new_config.opendrive_file);
    this->angle_tol=new_config.angle_tol;
    this->dist_tol=new_config.dist_tol;
    this->multi_hyp=new_config.multi_hyp;
    this->config=new_config;
    this->roadmap.set_cost_type((cost_t)new_config.cost_type);
  }catch(CException &e){
    ROS_ERROR_STREAM(e.what());
  }
}

void OpendriveGlobalPlanner::clearRobotCell(const geometry_msgs::PoseStamped& global_pose, unsigned int mx, unsigned int my) 
{
  if (!initialized_) 
  {
    ROS_ERROR("This planner has not been initialized yet, but it is being used, please call initialize() before use");
    return;
  }

  //set the associated costs in the cost map to be free
  costmap_->setCost(mx, my, costmap_2d::FREE_SPACE);
}

bool OpendriveGlobalPlanner::makePlanService(nav_msgs::GetPlan::Request& req, nav_msgs::GetPlan::Response& resp) 
{
  makePlan(req.start, req.goal, resp.plan.poses);

  resp.plan.header.stamp = ros::Time::now();
  resp.plan.header.frame_id = frame_id_;

  return true;
}

void OpendriveGlobalPlanner::mapToWorld(double mx, double my, double& wx, double& wy) 
{
  wx = costmap_->getOriginX() + mx * costmap_->getResolution();
  wy = costmap_->getOriginY() + my * costmap_->getResolution();
}

bool OpendriveGlobalPlanner::worldToMap(double wx, double wy, double& mx, double& my) 
{
  double origin_x = costmap_->getOriginX(), origin_y = costmap_->getOriginY();
  double resolution = costmap_->getResolution();

  if (wx < origin_x || wy < origin_y)
    return false;

  mx = (wx - origin_x) / resolution;
  my = (wy - origin_y) / resolution;

  if (mx < costmap_->getSizeInCellsX() && my < costmap_->getSizeInCellsY())
    return true;

  return false;
}

bool OpendriveGlobalPlanner::makePlan(const geometry_msgs::PoseStamped& start, const geometry_msgs::PoseStamped& goal,std::vector<geometry_msgs::PoseStamped>& plan) 
{
  return makePlan(start, goal, 0.0, plan);
}

bool OpendriveGlobalPlanner::makePlan(const geometry_msgs::PoseStamped& start, const geometry_msgs::PoseStamped& goal,double tolerance, std::vector<geometry_msgs::PoseStamped>& plan) 
{
  double yaw,best_cost;
  unsigned int best_path_index;
  std::vector<unsigned int> best_path;
  std::vector<double> x,y,heading;
  std::vector< std::vector<unsigned int> > paths;
  std::vector<double> costs;
  std::vector<TMapPose> start_candidates,end_candidates;

  boost::mutex::scoped_lock lock(mutex_);
  if (!initialized_) {
    ROS_ERROR("This planner has not been initialized yet, but it is being used, please call initialize() before use");
    return false;
  }

  //clear the plan, just in case
  plan.clear();

  ros::NodeHandle n;
  std::string global_frame = frame_id_;

  //until tf can handle transforming things that are way in the past... we'll require the goal to be in our global frame
  if (goal.header.frame_id != global_frame) 
  {
    ROS_ERROR("The goal pose passed to this planner must be in the %s frame.  It is instead in the %s frame.", global_frame.c_str(), goal.header.frame_id.c_str());
    return false;
  }

  if (start.header.frame_id != global_frame) 
  {
    ROS_ERROR("The start pose passed to this planner must be in the %s frame.  It is instead in the %s frame.", global_frame.c_str(), start.header.frame_id.c_str());
    return false;
  }

/*
  double wx = start.pose.position.x;
  double wy = start.pose.position.y;

  unsigned int start_x_i, start_y_i, goal_x_i, goal_y_i;
  double start_x, start_y, goal_x, goal_y;

  if (!costmap_->worldToMap(wx, wy, start_x_i, start_y_i)) 
  {
    ROS_WARN("The robot's start position is off the global costmap. Planning will always fail, are you sure the robot has been properly localized?");
    return false;
  }
  worldToMap(wx, wy, start_x, start_y);

  //clear the starting cell within the costmap because we know it can't be an obstacle
  clearRobotCell(start, start_x_i, start_y_i);

  int nx = costmap_->getSizeInCellsX(), ny = costmap_->getSizeInCellsY();
*/
  try{
    ROS_WARN("Make Plan");
    yaw=tf2::getYaw(start.pose.orientation);
    this->roadmap.set_start_pose(start.pose.position.x,start.pose.position.y,yaw,this->dist_tol,this->angle_tol);
    this->roadmap.get_start_pose_candidates(start_candidates);
    std::cout << "Start pose candidates: " << std::endl;
    for(unsigned int i=0;i<start_candidates.size();i++)
    {
      std::cout << i << ". x: " << start_candidates[i].pose.x << ", y:" << start_candidates[i].pose.y << ", heading: " << start_candidates[i].pose.heading << std::endl;
      std::cout << "   road map node: " << start_candidates[i].node_index << " child: " << start_candidates[i].child_index << std::endl;
      std::cout << "   distance from desired pose: " << start_candidates[i].dist << std::endl;
    }
    yaw=tf2::getYaw(goal.pose.orientation);
    this->roadmap.set_end_pose(goal.pose.position.x,goal.pose.position.y,yaw,this->dist_tol,this->angle_tol);
    this->roadmap.get_end_pose_candidates(end_candidates);
    std::cout << "End pose candidates: " << std::endl;
    for(unsigned int i=0;i<end_candidates.size();i++)
    {
      std::cout << i << ". x: " << end_candidates[i].pose.x << ", y:" << end_candidates[i].pose.y << ", heading: " << end_candidates[i].pose.heading << std::endl;
      std::cout << "   road map node: " << end_candidates[i].node_index << " child: " << end_candidates[i].child_index << std::endl;
      std::cout << "   distance from desired pose: " << end_candidates[i].dist << std::endl;
    }
    this->roadmap.find_shortest_path(this->multi_hyp);
    this->roadmap.get_paths(costs,paths);
    for(unsigned int i=0;i<paths.size();i++)
    {
      std::cout << "path " << i << ":" << std::endl;
      for(unsigned int j=0;j<paths[i].size();j++)
        std::cout << paths[i][j] << ",";
      std::cout << std::endl;
    }
    best_path_index=this->roadmap.get_best_path(best_cost,best_path);
    std::cout << "best path index: " << best_path_index << std::endl;
    for(unsigned int i=0;i<best_path.size();i++)
      std::cout << best_path[i] << ",";
    std::cout << std::endl;
    this->roadmap.get_trajectory(best_path_index,x,y,heading);
    plan.resize(x.size());
    ros::Time plan_time=ros::Time::now();
    for(unsigned int i=0;i<x.size();i++)
    {
      plan[i].header.frame_id=frame_id_;
      plan[i].header.stamp=plan_time;
      plan[i].pose.position.x=x[i];
      plan[i].pose.position.y=y[i];
      plan[i].pose.position.y=y[i];
      tf2::Quaternion orientation;
      orientation.setRPY(0.0,0.0,heading[i]);
      plan[i].pose.orientation=tf2::toMsg(orientation);
    }
  }catch(CException &e){
    ROS_ERROR_STREAM(e.what());
    return false;
  }
  //make sure to resize the underlying array that Navfn uses

  //publish the plan for visualization purposes
  publishPlan(plan);
  return !plan.empty();
}

void OpendriveGlobalPlanner::publishPlan(const std::vector<geometry_msgs::PoseStamped>& path) 
{
  if (!initialized_) 
  {
    ROS_ERROR("This planner has not been initialized yet, but it is being used, please call initialize() before use");
    return;
  }

  //create a message for the plan
  nav_msgs::Path gui_path;
  gui_path.poses.resize(path.size());

  gui_path.header.frame_id = frame_id_;
  gui_path.header.stamp = ros::Time::now();

  // Extract the plan in world co-ordinates, we assume the path is all in the same frame
  for (unsigned int i = 0; i < path.size(); i++) 
  {
    gui_path.poses[i] = path[i];
  }

  plan_pub_.publish(gui_path);
}

} //end namespace iri_opendrive_global_planner
